  '******************************************************************************
  '* LoRa_TX_SX1278_FN3.bas
  '*
  '* Originalet hittat från nätet 18.12.2017:
  '* https://www.dropbox.com/sh/wvspej3l4x4bq1e/AACz_3Pk7OOhwImUeYWMsp08a/MicroMite%20MK2%20LoRa%20Programs?dl=0
  '*
  '* Måste ändra de använda pinnarna för att passa µMite Backpack.
  '*
  '* Originalet: Micromite MK2 LoRa TX V2.BAS
  'Copyright of the author Stuart Robinson - 26/04/15 12:00
  'This program may be used free of charge for personal, recreational and educational purposes only.
  'This program, or parts of it, may not be used for or in connection with any commercial purpose without the explicit
  'permission of the author Stuart Robinson.
  
  '                            ***********    ***     ***
  '                            ***********     ***   ***
  '                                ***          *** ***
  '                                ***           *****
  '                                ***            ***
  '                                ***           *****
  '                                ***          *** ***
  '                                ***         ***   ***
  '                                ***        ***     ***
  'AiThinker Lora Ra-01:
  '
  '   LoRa modul          µMite 28pin
  '   GND                 GND
  '   3.3V                3.3V
  '   RESET               pin 9 (se nedan LORA.PReset = 9)
  '   NSS (Chip Select)   pin 4 (se nedan LORA_PNSS = 4)
  '   MOSI                MOSI pin 3
  '   MISO                MISO pin 14
  '   SCK                 pin 25 SPI Clk
  
  option autorun on
  OPTION DEFAULT NONE
  OPTION EXPLICIT
  
  'initialise symbols and variables
  PRINT "LORA.Init"
  
  'Variables
  DIM INTEGER LORA_NULL 'För att ta emot nonsens data som inte behöver användas.
  DIM INTEGER LORA_Data
  DIM INTEGER LORA_Reg
  DIM INTEGER LORA_TXBUFF(256)
  DIM INTEGER LORA_RXBUFF(256)
  DIM INTEGER LORA_RXStart
  DIM INTEGER LORA_RXEnd
  DIM INTEGER LORA_TXStart
  DIM INTEGER LORA_TXEnd
  DIM INTEGER LORA_CMDBUFF(32)
  DIM INTEGER LORA_FCRCerror
  DIM INTEGER LORA_FRXTimeout
  DIM INTEGER LORA_FTXOK
  DIM INTEGER LORA_FRXOK
  DIM INTEGER LORA_RXPacketType
  DIM INTEGER LORA_TXPacketType
  DIM INTEGER LORA_TXDestination
  DIM INTEGER LORA_TXSource
  DIM INTEGER LORA_RXDestination
  DIM INTEGER LORA_RXSource
  DIM INTEGER LORA_BackGroundRSSI
  DIM INTEGER LORA_PacketRSSI
  DIM INTEGER LORA_PacketSNR
  DIM INTEGER LORA_RXpacketCount
  DIM INTEGER LORA_CRCerrorcount
  DIM INTEGER LORA_TXpacketCount
  DIM INTEGER LORA_TXPacketL
  DIM INTEGER LORA_RXPacketL
  DIM INTEGER LORA.FDeviceError
  DIM FLOAT LORA.Frequency
  
  
  'Pins
  CONST LORA_PNSS = 4     'pin number where the NSS line for the LoRa device is connected.
  CONST LORA_PLED1 = 5     'Pin number for LED1
  CONST LORA.PReset = 9    'Pin where LoRa device reset line is connected
  CONST LORA_PPWMCH = 2     'Channel number for PWM direct MOD, used for FM tones.
  CONST ARS.RTCAdr = &H51   'Address of PCF8563 (För serial interface till LCD, behöver inte denna.)
  CONST ARS.EEAdr = &H50    'Address of EEPROM
  
  
  'LORA names for bandwidth settings. Sid 112 i datasheet.
  CONST LORA_BW7_8 = 0      '7.8khz
  CONST LORA_BW10_4 = 16    '10.4khz
  CONST LORA_BW15_6 = 32    '15.6khz
  CONST LORA_BW20_8 = 48    '20.8khz
  CONST LORA_BW31_2 = 64    '31.2khz
  CONST LORA_BW41_7 = 80    '41.7khz
  CONST LORA_BW62_5 = 96    '62.5khz
  CONST LORA_BW125 = 112    '125khz
  CONST LORA_BW250 = 128    '250khz
  CONST LORA_BW500 = 144    '500khz
  
  'Spreading Factor. Sid 27 i datasheet.
  CONST LORA_SF6 = 6
  CONST LORA_SF7 = 7
  CONST LORA_SF8 = 8
  CONST LORA_SF9 = 9
  CONST LORA_SF10 = 10
  CONST LORA_SF11 = 11
  CONST LORA_SF12 = 12
  
  'LORA names for coding rate settings. Sid 27.
  CONST LORA_CR4_5 = 2      '4:5
  CONST LORA_CR4_6 = 4      '4:6
  CONST LORA_CR4_7 = 6      '4:7
  CONST LORA_CR4_8 = 8      '4:8
  
  'LORA Header Settings. Sid 30.
  CONST LORA_EXPLICIT    = 0  'Use to set explicit header
  CONST LORA_IMPLICIT    = 1  'Use to set implicit header
  
  CONST LORA_RX = 0           'used to define array dimension for RX
  CONST LORA_TX = 1           'used to define array dimension for RX
  
  'SX1278 Register names. Sid 90. Registernamnen enligt LoRa-Mode.
  CONST LORA_RegFifo = &H00
  CONST LORA_WRegFifo = &H80  'Write and Read bit (vnr bit)=1 for write. Sid 80.
  CONST LORA_RegOpMode = &H01
  CONST LORA_RegFdevLsb = &H05
  CONST LORA_RegFrMsb = &H06  'Sid 108
  CONST LORA_RegFrMid = &H07  'Sid 109
  CONST LORA_RegFrLsb = &H08  'Sid 109
  CONST LORA_RegPaConfig = &H09 'sid 109
  CONST LORA_RegOcp = &H0B
  CONST LORA_RegLna = &H0C
  CONST LORA_RegFifoAddrPtr = &H0D
  CONST LORA_RegFifoTxBaseAddr = &H0E
  CONST LORA_RegFifoRxBaseAddr = &H0F
  CONST LORA_RegFifoRxCurrentAddr = &H10
  CONST LORA_RegIrqFlagsMask = &H11
  CONST LORA_RegIrqFlags = &H12
  CONST LORA_RegRxNbBytes = &H13
  CONST LORA_RegRxHeaderCntValueMsb = &H14
  CONST LORA_RegRxHeaderCntValueLsb = &H15
  CONST LORA_RegRxPacketCntValueMsb = &H16
  CONST LORA_RegRxPacketCntValueLsb = &H17
  CONST LORA_RegPktSnrValue = &H19
  CONST LORA_RegPktRssiValue = &H1A
  CONST LORA_RegRssiValue = &H1B
  CONST LORA_RegFsiMSB = &H1D
  CONST LORA_RegFsiLSB = &H1E
  CONST LORA_RegModemConfig1 = &H1D
  CONST LORA_RegModemConfig2 = &H1E
  CONST LORA_RegSymbTimeoutLsb = &H1F
  CONST LORA_RegPreambleLsb = &H21
  CONST LORA_RegPayloadLength = &H22
  CONST LORA_RegFifoRxByteAddr = &H25
  CONST LORA_RegModemConfig3 = &H26
  CONST LORA_RegPacketConfig2 = &H31
  CONST LORA_TXdefaultpower = 10
  CONST LORA_Deviation = &H52
  CONST LORA_LowDoptON = &H08             'value to turn low data rate optimisation on
  CONST LORA_LowDoptOFF = &H00            'value to turn low data rate optimisation off
  CONST LORA_PrintASC = 0                 'value to cause buffer print to appear as ASCII
  CONST LORA_PrintNum = 1                 'value to cause buffer print to appear as decimal numbers
  CONST LORA_PrintHEX = 2                 'value to cause buffer print to appear as hexeecimal numbers
  
  DIM DisplayString as string   'Text som ska visas på LCD display
  DIM String.To.Send as string  'Text som ska sändas
  
  dim as integer Blinker      'Blinker to show that program is running
  
  PRINT "Micromite II LoRa TX V2.BAS"
  PRINT "Stuart Robinson - April 2015"
  Print "Fredrik Nissfolk - December 2017"
  
  cls
  text 10,10,"Micromite LoRa TX","LT",1,2,rgb(yellow)
text 10,70,"Fredrik Nissfolk - maj 2018","LT"
pause 5000

  
  '***************************************************************************
'INSTÄLLNINGAR:
  DIM as Float Frequency=434.4  'MHz 410...525
  DIM as integer TX_Power=2 '2...17dBm
  DIM as integer SpreadingFactor=LORA_SF10  '6...12
  DIM as integer CodingRate=LORA_CR4_7     '4_5, 4_6, 4_7 or 4_8
  dim as integer BandWidth=LORA_BW41_7     '(kHz) 7_8, 10_4, 15_6, 20_8, 31_2, 41_7, 62_5, 125, 250 or 500
  dim as integer TX_timeout=10          '(seconds)High Spreadingfactor and high CodingRate need high TX_timeout.
  '***************************************************************************
  print "CodingRate:"CodingRate
  Print "SpreadingFactor:"SpreadingFactor
  Print "BandWidth"BandWidth
  
Main:
  CLS
  text 10,0,"LoRa Transmitter", "LT",1,1,rgb(yellow)
  text 10,20,"Modulation: CR="+str$(CodingRate)+" SF="+str$(SpreadingFactor)+" BW="+str$(BandWidth),"LT"
  Program.Setup
  
LOOP1:
  
  LORA.Setup                                  'Do the initial LoRa Setup
  
  IF LORA.Checkdevice() = 0 then
    PRINT "ERROR, Device not Working"
    PRINT ""
    TEXT 160,0,"                         ","LT"
    TEXT 160,0,"ERROR, Device not Working","LT"
    PAUSE 2000
    GOTO LOOP1
  ELSe
    PRINT "Device OK"
    TEXT 160,0,"       Device OK         ","LT"
  END IF
  
  LoRa.SeTFreq(Frequency)                      'Set the frequency in Mhz
  
  PRINT "Programmed Frequency " LORA.GetFreq() "Mhz"     'Use the function, get the frequency programmed
  DisplayString="Programmed Frequency "+str$(LORA.GetFreq())+" MHz"
  TEXT 10,40,DisplayString,"LT"
  LORA.Print                                 'Print all the device registers
  LORA.Tone(1000,1000,2)                      'Transmit a LoRa tone
  LORA.SetModem(BandWidth,SpreadingFactor,CodingRate,LORA_EXPLICIT,LORA_LowDoptOFF)  'Setup the LoRa modem parameters
  ' Modem parameters:
  '   BW=41.7 kHz, Spreading factor= 8, Coding rate= 4:5, Explicit header=0, LowDataRateOpt=OFF
  '   Highest sensitivity:
  '     -Max Spreading factor= SF12
  '     -Codingrate=4/8
  '     -High bandwidth
  LORA.PrintModem  'Print the modem parameters
  LORA.FillTX
  
  do
    'LORA.FillTX                                             'Fill transmit buffer with test data; ABCDEFGH.....
    TX_Power=TX_Power+1
    IF TX_Power>17 then TX_Power=2:end if
    String.To.Send="TX power="+str$(TX_power)+"dBm  "+date$+"  "+time$+"A1234567890B1234567890C1234567890"
    String.To.Send=String.To.Send+"D1234567890E1234567890F1234567890"
    'FN_FillTX(TX_string_FN as string)
    FN_FillTX(String.To.Send)
    'LORA.Send(TXStart,TXEnd,TXPacketType,TXDestination,TXSource,TXTimeout,TXPower):
    '   TXStart;  Start  Location  in  buffer,  pointer  to  start  of  packet  in  TXbuffer
    '   TXEnd;  End  Location  in  buffer,  pointer  to  end  of  packet  in  Txbuffer
    '   PacketType,  0-255,  byte  defining  the  type  of  packet
    '   TXDestination;  TX  Destination  Node,  0-255,  Destination  node  for  packet,  255  is  broadcast
    '   TXSource;  Source  Node,  0-255,  Node  packet  is  sent  from
    '   TXTimeout;  Transmitter  Time-out  in  mS,  0  for  no  time-out  check
    '   TXPower;  Transmit  power  in  dBm,  17  to  2
    LORA.Send(LORA_TXStart,LORA_TXEnd,32,255,1,TX_timeout,TX_Power)       'send the packet
    LORA.TXPKTInfo                                          'print packet inforamtion
    LORA.TXBuffPrint(LORA_PrintASC)
    if Blinker<1 then 
    blinker=1
    text 0,0,"*","LT" 
    else 
    Blinker=0
    text 0,0," ","LT"
    end if
    PAUSE 3000                                               'print decimal numbers of packet sent
  loop
  
END
  
  
SUB Program.Setup
  PRINT "Program.Setup"
  SETPIN LORA_PLED1, DOUT
  SETPIN LORA.PReset, DOUT
  PIN(LORA.PReset) = 1
  SETPIN LORA_PNSS, DOUT
  PIN(LORA_PNSS) = 1
  'Originalprogrammet är gjort för I2C displayer. Om ILI9341 används så kan SPI inte vara öppen hela tiden.
  'Måste öppna och stänga SPI vid varje Read och Write för att SPI inte ska kollidera med LCD-touch SPI.
  'Alternativt så kan också en Parallell (HD44780 16x2 eller 20x4) LCD användas.
  '****** TA BORT SPI OPEN
  'SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  '******
END SUB
  
SUB LORA.Setup
  'initialise LoRa device registers and check its responding
  PRINT "LORA.Setup"
  LORA.ResetDev'Clear all registers to default
  LORA.Write(LORA_RegOpMode,&H08)           'RegOpMode, need to set to sleep mode before configure for LoRa mode. Sid 108.
  LORA.Write(LORA_RegOcp,&H0B)              'RegOcp. Sid 109: Overload current Protection OFF. Imax=100mA.
  LORA.Write(LORA_RegLna,&H23)              'RegLna. Sid 110: G1=max gain, Default L.F. LNA current, H.F. LNA Boost
  LORA.Write(LORA_RegSymbTimeoutLsb,&HFF)   'RegSymbTimeoutLsb. Sid 113:
  LORA.Write(LORA_RegPreambleLsb,&H0C)      'RegPreambleLsb, default. Sid 113: Preamble lenghth=12. Default är 12
  LORA.Write(LORA_RegFdevLsb,LORA_Deviation)'LSB of deviation, 5kHz= default. Sid 93.
END SUB
  
FUNCTION LORA.Checkdevice() AS INTEGER
  'check there is a device out there
  PRINT "LORA.Checkdevice"
  LOCAL INTEGER LORA_LVAR1, LORA_LVAR2,LORA_LVAR3
  LORA.Write(LORA_RegFrMSB,&H5A)  'testar med att sätta frekvensen till 362.66632MHz
  LORA.Write(LORA_RegFrMid,&HAA)
  LORA.Write(LORA_RegFrLSB,&HA5)
  'Now Check to see if the LoRa device has programmed correctly, check the 3 frequency setting registers.
  
  SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  PIN(LORA_PNSS) = 0
  LORA_LVAR1 = SPI(LORA_RegFrMsb)     'write out the first frequency register to read
  LORA_LVAR1 = SPI(0)                 'Read RegFrMSB
  LORA_LVAR2 = SPI(0)                 'Read RegFrMid
  LORA_LVAR3 = SPI(0)                 'Read RegFrLSB
  PIN(LORA_PNSS) = 1
  SPI close
  
  if LORA_LVAR1  <> &H5A OR LORA_LVAR2 <> &HAA OR LORA_LVAR3 <> &HA5 then
    LORA.Checkdevice = 0  'hamnar hit om frekvensregistren inte avlästs korrekt.
    LORA.FDeviceError = 0
  else
    LORA.Checkdevice = 1
    LORA.FDeviceError = 1
  end if
END FUNCTION
  
SUB LORA.ResetDev
  PRINT "LORA.ResetDev"
  PIN(LORA.PReset) = 0
  PAUSE 1
  PIN(LORA.PReset) = 1
  PAUSE 10
END SUB
  
SUB LORA.Write(LORA_LReg AS INTEGER,LORA_LData AS INTEGER) AS INTEGER
  'PRINT "LORA.Write"
  'PRINT "Writing ";
  'PRINT LORA_RegL;
  'PRINT " ";
  'PRINT LORA_DataL
  LORA_LReg = LORA_LReg OR &H80
  
  SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  PIN(LORA_PNSS) = 0
  LORA_NULL = SPI(LORA_LReg)                      'send register address for a read 0-&7F
  LORA_NULL = SPI(LORA_LData)                     'and read it in, data is returned in LORA_Reg
  PIN(LORA_PNSS) = 1
  spi close
END SUB
  
FUNCTION LORA.Read(LORA_LReg AS INTEGER) AS INTEGER   'uses LORA_Reg to get regdata
  'PRINT "LORA.Read"
  SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  PIN(LORA_PNSS) = 0
  LORA.Read = SPI(LORA_LReg)                      'send register address for a read 0-&7F
  LORA.Read = SPI(0)                              'and read it in, data is returned in LORA_Reg
  PIN(LORA_PNSS) = 1
  spi close
END FUNCTION
  
SUB LoRa.SetFreq(LORA_LFreq As FLOAT)
  '*** Beräknar registervärden baserat på önskad frekvens och sänder till SX1278.
  LOCAL INTEGER LORA_LFMSB, LORA_LFMid, LORA_LFLSB, LORA_LINTFreq
  LORA_LINTFreq = INT(((LORA_LFreq * 1000000)/61.03515625)) 'Sid 109: 61.035. Tydligen justerat enligt xtallens riktiga frekvens.
  PRINT "LoRa.SETFreq"
  PRINT "&H" HEX$(LORA_LINTFreq,6)
  LORA_LFMSB =  LORA_LINTFreq >> 16
  LORA_LFMid =  (LORA_LINTFreq AND &H00FF00) >> 8
  LORA_LFLSB =  LORA_LINTFreq AND &H0000FF
  LORA.Write(LORA_RegFrMSB,LORA_LFMSB)
  LORA.Write(LORA_RegFrMid,LORA_LFMid)
  LORA.Write(LORA_RegFrLSB,LORA_LFLSB)
END SUB
  
FUNCTION LORA.GetFreq() AS FLOAT
  '*** Beräknar frekvens baserat på registervärden.
  'PRINT "LORA.GetFreq"
  LOCAL INTEGER LORA_LFMSB, LORA_LFMid, LORA_LFLSB
  LORA_LFMSB = LORA.READ(LORA_RegFrMSB)
  LORA_LFMid = LORA.READ(LORA_RegFrMid)
  LORA_LFLSB = LORA.READ(LORA_RegFrLSB)
  LORA.GetFreq = (LORA_LFMSB * &H10000) +  (LORA_LFMid * &H100) + (LORA_LFLSB)
  LORA.GetFreq = (LORA.GetFreq * 61.03515625) / 1000000
  LORA.Frequency = LORA.GetFreq   'So we have the value avaialble outside of routine
END FUNCTION
  
SUB LORA.SetModem(LORA_LBW AS INTEGER, LORA_LSF AS INTEGER, LORA_LCR   AS INTEGER, LORA_LHDR AS INTEGER, LORA_LLDROPT AS INTEGER)
  'setup the LoRa modem parameters
  PRINT "LORA.SetModem  "
  LOCAL INTEGER LORA_LVAR1, LORA_LVAR2
  LORA_LVAR1 = LORA_LBW + LORA_LCR + LORA_LHDR          'calculate value of RegModemConfig1. LORA_LHDR är inte deklarerat.
  LORA_LVAR2 = LORA_LSF * 16 + 7                        'calculate value of RegModemConfig2, ($07; Header indicates CRC on, RX Time-Out MSB = 11
  LORA.Write(LORA_RegOpMode,&H08)                       'RegOpMode, need to set to sleep mode before configure for LoRa mode
  LORA.Write(LORA_RegOpMode,&H88)                       'goto LoRa mode
  LORA.Write(LORA_RegModemConfig1,LORA_LVAR1)   'Sid 112: BW, CodingRate, HeaderMode
  LORA.Write(LORA_RegModemConfig2,LORA_LVAR2)   'Sid 113: Spreading Factor, TxContMode=single packet, CRC=ON, RxTimeoutMSB=0x11
  LORA.Write(LORA_RegModemConfig3,LORA_LLDROPT) 'Sid 114: LORA_LLDROPT är inte deklarerad. Default för RegModemConfig3 är 0.
END SUB
  
SUB LORA.PrintModem
  PRINT "LORA.PrintModem "
  PRINT " RegModemConfig1 (reg 0x1D): &H" HEX$(LORA.Read(LORA_RegModemConfig1),2);" 0b";bin$(LORA.Read(LORA_RegModemConfig1),8)
  PRINT " ";
  PRINT " RegModemConfig2 (reg 0x1E): &H" HEX$(LORA.Read(LORA_RegModemConfig2),2);" 0b";bin$(LORA.Read(LORA_RegModemConfig2),8)
  PRINT " ";
  PRINT " RegModemConfig3 (reg 0x26): &H" HEX$(LORA.Read(LORA_RegModemConfig3),2);" 0b";bin$(LORA.Read(LORA_RegModemConfig3),8)
END SUB
  
  PRINT hex$(LORA_LLoopv1,1);
  
SUB LORA.Print
  'prints the contents of LORA registers to serial terminal
  LOCAL INTEGER LORA_LLoopv1,LORA_LLoopv2
  PRINT "LORA.Print"
  PRINT ""
  PRINT "Reg  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F"
  LORA_Reg = 0                                               'point to first register of LORA Device
  for LORA_LLoopv1 = 0 to 7                                  '8 lines
    PRINT hex$(LORA_LLoopv1,1);
    PAUSE 2
    PRINT "0  ";
    PAUSE 2
    for LORA_LLoopv2 = 0 to 15                           '16 bytes per line
      PRINT hex$(LORA.Read(LORA_Reg),2);
      PRINT " ";
      LORA_Reg = LORA_Reg + 1
    next LORA_LLoopv2
    PRINT ""
  next LORA_LLoopv1
  PRINT ""
END SUB
  
  '************************
  'START LORA Transmit Stuff
  '************************
SUB LORA.Send(LORA_LTXStart AS INTEGER,LORA_LTXEnd AS INTEGER,LORA_LTXPacketType AS INTEGER,LORA_LTXDestination AS INTEGER,LORA_LTXSource AS INTEGER,LORA_LTXTimeout as INTEGER,LORA_LTXPower AS INTEGER)
  'fills FIFO with 3 header bytes, then from LORA_TXBUFF(256) starting at LORA_TXStart ending at LORA_TXEnd,maximum of 252 btes
  PRINT "LORA.Send " LORA_LTXStart LORA_LTXEnd
  Print "Transmit Timeout"  LORA_LTXTimeout "Seconds"
  LOCAL INTEGER LORA_LVAR1, LORA_LChar, LORA_LTXPacketL
  Timer=0
  LORA_TXStart = LORA_LTXStart
  LORA_TXEnd = LORA_LTXEnd
  LORA_TXDestination = LORA_LTXDestination
  LORA_TXPacketType = LORA_LTXPacketType
  LORA_TXSource = LORA_LTXSource
  LORA.Write(LORA_RegOpMode,&H09)           'Sätter i Standby mode. Sid 108
  LORA.Write(LORA_RegIrqFlags,&HFF)         'Sid 111: Clear the IRQ.
  LORA.Write(LORA_RegIrqFlagsMask,&HF7)     'Sid 110: Maskerar alla andra än TxDone, som kommer att ge IRQ när TX är klart.
  LORA.Write(LORA_RegFifoTxBaseAddr,&H80)
  LORA.Write(LORA_RegFifoAddrPtr,&H80)                      'start burst read
  
  SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  PIN(LORA_PNSS) = 0
  LORA_NULL = SPI(LORA_WRegFifo)                            'send register address for a burst write
  LORA_NULL = SPI(LORA_LTXPacketType)                       'Write the packet type
  LORA_NULL = SPI(LORA_LTXDestination)                      'Destination node
  LORA_NULL = SPI(LORA_LTXSource)                           'Source node
  LORA_LTXPacketL = 3
  for LORA_LVAR1 = LORA_TXStart to LORA_TXEnd
    LORA_LTXPacketL = LORA_LTXPacketL + 1
    if LORA_LTXPacketL > 252 then                       'check for overlong packet here, wraps around from limit at 251 to 0
      PRINT "**** ERROR,PacketatLimit ****"
      EXIT SUB
    end if
    LORA_LChar = LORA_TXBUFF(LORA_LVAR1)
    LORA_NULL = SPI(LORA_LChar)
  next LORA_LVAR1
  PRINT ""
  LORA_TXPacketL = LORA_LTXPacketL
  PIN(LORA_PNSS) = 1                                        'finish the burst write
  spi close
  LORA.Write(LORA_RegPayloadLength,LORA_LTXPacketL)
  LORA.TXonLoRa(LORA_LTXPower)
  LORA_LVAR1 = 0
  'LORA_LTXTimeout = LORA_LTXTimeout * 100                   'convert seconds to mS
  
  do
    pause 10
    LORA_LVAR1 = LORA_LVAR1 + 1
    LORA_LChar = LORA.Read(LORA_RegIrqFlags) AND &H08     'mask off the TX done flag
  loop until LORA_LChar > 0 OR TIMER > LORA_LTXTimeout*1000   'use a timeout counter, just in case the TX sent flag is missed
  
  LORA.TXoff
  if Timer > LORA_LTXTimeout*1000 then
    PRINT " **** ERROR,TX timeout ****"
    text 10,120,"**** ERROR,TX timeout ****","LT"
    LORA_FTXOK = 0
  ELSE
    LORA_FTXOK = 1
    LORA_TXpacketCount = LORA_TXpacketCount + 1
  end if
  text 10,215,"                            ","LT"
  text 10,215,"---TX needed "+str$(TIMER)+" ms","LT"
END SUB
  
SUB LORA.FillTX
  PRINT "LORA.FillTX"
  LOCAL INTEGER LORA_LLOOP
  LORA_TXStart = 0
  LORA_TXEnd = 0
  FOR LORA_LLOOP = 65 to 90  'ASCII tecken
    LORA_TXBUFF(LORA_TXEnd) =  LORA_LLOOP
    LORA_TXEnd = LORA_TXEnd + 1
  NEXT LORA_LLOOP
  LORA_TXEnd = LORA_TXEnd - 1
end SUB
  
Sub FN_FillTX(TX_string_FN as string)
  'Fyller TX buffern med text.
  'Text som ska TXas finns i TX_string_FN
  local as integer String.Len=0  'String length
  LOCAL INTEGER LORA_LLOOP
  print "FN_fillTX";TX_string_FN
  LORA_TXStart = 0
  LORA_TXEnd = 0                  'Index starts from 0, but character numbering from 1
  String.Len=LEN(TX_string_FN)-1
  FOR LORA_LLOOP = LORA_TXStart to String.Len  'ASCII tecken
    LORA_TXBUFF(LORA_LLOOP) =  asc(MID$(TX_string_FN, LORA_LLOOP+1,1)) 'take one character from the string
    LORA_TXEnd = LORA_TXEnd + 1
  NEXT LORA_LLOOP
  LORA_TXEnd = LORA_TXEnd - 1   'TXEnd got one step too much.
  print "FN_FillTX  Start=";LORA_TXStart;"  End=";LORA_TXEnd;" LEN=";LEN(TX_string_FN)
End sub
  
SUB LORA.TXPKTInfo
  PRINT "LORA.TXPKTInfo  "
  PRINT "TXtype," LORA_TXPacketType;
  PRINT ",TXDestination," LORA_TXDestination;
  PRINT ",TXSource," LORA_TXSOURCE;
  PRINT ",TXPacketLength," LORA_TXPacketL;
  PRINT ",TXPacketCount," LORA_TXpacketCount
END SUB
  
  
SUB LORA.TXBuffPrint(LORA_LPrint as INTEGER)
  PRINT "LORA.TXBuffPrint ";
  LOCAL INTEGER LORA_LLoop, LORA_LData
  DisplayString=""
  for LORA_LLoop = LORA_TXStart to LORA_TXEnd
    LORA_LData = LORA_TXBUFF(LORA_LLoop)
    IF LORA_LPrint = 0 then
      PRINT CHR$(LORA_LData);
      DisplayString=DisplayString+CHR$(LORA_LData)
    ELSE IF LORA_LPrint = 1 then
      PRINT LORA_LData;
      DisplayString=DisplayString+LORA_LData
    ELSE IF LORA_LPrint = 2 then
      PRINT HEX$(LORA_LData, 2) " ";
      DisplayString=DisplayString+hex$(LORA_LData,2)+" "
    END IF
  Next LORA_LLoop
  PRINT ""
  
  text 10,100,"Transmitted Package:","LT"
  text 10,120,"                                       ","LT"
  text 10,120,MID$(DisplayString,1,35),"LT",1,1,rgb(blue), RGB(white)
  text 10,140,MID$(DisplayString,36,35),"LT",1,1,rgb(blue), RGB(white)
  text 10,160,MID$(DisplayString,71,35),"LT",1,1,rgb(blue), RGB(white)
  text 10,180,MID$(DisplayString,106,35),"LT",1,1,rgb(blue), RGB(white)
end SUB
  
SUB LORA.Tone(LORA_LFreq AS INTEGER,LORA_LToneLen AS INTEGER,LORA_LTXPower AS INTEGER )
  'Transmit an FM tone
  PRINT "LORA.Tone"
  LORA.DirectSetup
  LORA.TXONDirect(LORA_LTXPower)                        'TXON, power 10
  LORA.Write(LORA_RegFdevLsb,LORA_Deviation)            'We are generating a tone so set the deviation, 5kHz
  PWM LORA_PPWMCH, LORA_LFreq, 50                       'start pwm output
  Pause LORA_LToneLen
  PWM LORA_PPWMCH, STOP
  LORA.TXOFF
END SUB
  
SUB LORA.DirectSetup
  PRINT "LORA.DirectSetup"
  LORA.Write(LORA_RegOpMode,&H08)
  LORA.Write(LORA_RegPacketConfig2,&H00)                'set continuous mode
END SUB
  
SUB LORA.TXONLoRa(LORA_LTXPower AS INTEGER)
  'turns on LoRa transmitter, Sends packet, power level is from 2 to 17
  PRINT "LORA.TXONLoRa" LORA_LTXPower "dBm"
  DisplayString="LoRa TX Power= "+str$(LORA_LTXPower)+"dBm"
  text 10,65,"                              ","LT"
  text 10,65,DisplayString,"LT"
  LOCAL INTEGER LORA_LVAR1
  LORA_LVAR1 = LORA_LTXPower + &HEE                     'has effect of converting 17 into 15
  LORA.Write(LORA_RegPaConfig,LORA_LVAR1)               'set TX power
  LORA.Write(LORA_RegOpMode,&H8B)             'TX on direct mode, low frequency mode
  PIN(LORA_PLED1) = 1                         'LoRa mode, Transmit mode.
END SUB
  
SUB LORA.TXONDirect(LORA_LTXPower AS INTEGER)
  'turns on transmitter,in direct mode for FSK and audio  power level is from 2 to 17
  PRINT "LORA.TXONDirect"
  LOCAL INTEGER LORA_LVAR1
  PRINT "Power " LORA_LTXPower
  LORA_LVAR1 = LORA_LTXPower + &HEE                      'has effect of converting 17 into 15
  LORA.Write(LORA_RegPaConfig,LORA_LVAR1)                'set TX power
  LORA.Write(LORA_RegOpMode,&H0B)                        'TX on direct mode, low frequency mode
END SUB
  
SUB LORA.TXOFF
  'turns off transmitter
  PRINT "LORA.TXOFF"
  LORA.Write(LORA_RegOpMode,&H08)                        'TX and RX to sleep, in direct mode
  PIN(LORA_PLED1) = 0
END SUB
  '************************
  'END LORA Transmit Stuff
  '************************
  
  '************************
  'START LORA Receive Stuff
  '************************
SUB LORA.Listen(LORA_LRXTimeout as INTEGER)
  'puts SX1278 into listen mode and receives packet
  'exits with packet in array LORA_RXBUFF(256)
  PRINT "LORA.Listen  ";
  LOCAL INTEGER LORA_LVAR1, LORA_LVAR2, LORA_LData, LORA_LLoop
  PRINT " Timeout";
  PRINT LORA_LRXTimeout;
  PRINT " Seconds"
  LORA_LRXTimeout = LORA_LRXTimeout * 220               'convert to number of 10mS steps
  LORA_FRXOK = 1                                        'Start assuming packet goes out OK, unless flag cleared
  LORA_FCRCerror = 0
  LORA_FRXTimeout = 0
  LORA_RXPacketType = 0
  LORA.Write(LORA_RegOpMode,&H09)
  LORA.Write(LORA_RegFifoRxBaseAddr,&H00)
  LORA.Write(LORA_RegFifoAddrPtr,&H00)
  LORA.Write(LORA_RegIrqFlagsMask,&H9F)
  LORA.Write(LORA_RegIrqFlags,&HFF)
  LORA.Write(LORA_RegOpMode,&H8D)
  LORA_BackGroundRSSI = LORA.Read(LORA_RegRssiValue)
  IF LORA_LRXTimeout = 0 THEN
    do
      LORA_LData = LORA.Read(LORA_RegIrqFlags)
      LORA_LVAR1 = LORA_LData AND &H40          'mask off RX in progress flag
    loop until LORA_LVAR1 > 0                       'wait until RX not in progress
  ELSE
    do
      LORA_LData = LORA.Read(LORA_RegIrqFlags)
      LORA_LVAR1 = LORA_LData AND &H40          'mask off RX in progress flag
      pause 1
      LORA_LVAR2 = LORA_LVAR2 + 1               'wait a bit
    loop until LORA_LVAR1 > 0 OR LORA_LVAR2 > LORA_LRXTimeout       'wait until RX not in progress
  END IF
  if LORA_LVAR2 > LORA_LRXTimeout then
    LORA_FRXOK = 0
    LORA_FRXTimeout = 1
    PRINT "Nothing Received"
END SUB
  end if
  PRINT "Packet Received "
  LORA_RXpacketCount = LORA_RXpacketCount + 1
  LORA_LVAR1 = LORA_LData AND &H20                      'now check for a CRC error
  LORA_RXPacketL = LORA.Read(LORA_RegRxNbBytes)
  LORA_PacketRSSI = LORA.Read(LORA_RegPktRssiValue)
  LORA_PacketSNR = LORA.Read(LORA_RegPktSnrValue)
  
  if LORA_LVAR1 > 0 then
    LORA_FRXOK = 0
    LORA_FCRCerror = 1
    PRINT "CRCFail !"
    LORA_CRCerrorcount = LORA_CRCerrorcount + 1
    return
  end if
  
  LORA_LData	=  LORA.Read(LORA_RegFifoRxCurrentAddr)   'read RX Fifo address of last packet
  LORA.Write(LORA_RegFifoRxCurrentAddr,LORA_LData)      'Set FIFO ptr
  
  SPI OPEN 10000000, 0, 8 'Öppnar SPI 10MHz, Mode 0, 8bit.
  PIN(LORA_PNSS) = 0                                    'start burst read
  LORA_LData = SPI(LORA_RegFifo)                        'send register address for a burst read 0-&7F
  LORA_RXPacketType = SPI(0)
  LORA_RXDestination = SPI(0)
  LORA_RXSource = SPI(0)                                'read the destination node the packet was sent to
  LORA_RXStart = 0
  LORA_RXEnd = LORA_RXPacketL - 4                       'adjust for destination,packettype and source bytes
  for LORA_LLoop = LORA_RXStart to LORA_RXEnd
    LORA_LData = SPI(0)
    LORA_RXBUFF(LORA_LLoop) = LORA_LData
  next LORA_LLoop
  PIN(LORA_PNSS) = 1                                    'Finish the Burst Read
  SPI END
  
END SUB
  
SUB LORA.RXBuffPrint(LORA_LPrint as INTEGER)
  PRINT "LORA.RXBuffPrint " LORA_RXStart " " LORA_RXEnd " Start>>";
  LOCAL INTEGER LORA_LLoop, LORA_LData
  for LORA_LLoop = LORA_RXStart to LORA_RXEnd
    LORA_LData = LORA_RXBUFF(LORA_LLoop)
    IF LORA_LPrint = 0 then
      PRINT CHR$(LORA_LData);
    ELSE IF LORA_LPrint = 1 then
      PRINT LORA_LData;
    ELSE IF LORA_LPrint = 2 then
      PRINT HEX$(LORA_LData, 2) " ";
    END IF
  Next LORA_LLoop
  PRINT "<<End"
end SUB
  
SUB LORA.RXPKTInfo
  LOCAL LORA_LVAR1 as INTEGER
  LOCAL LORA_LChar AS STRING
  PRINT "LORA.RXPKTInfo ";
  PRINT "RXtype," LORA_RXPacketType;
  PRINT ",Destination," LORA_RXDestination;
  PRINT ",Source," LORA_RXSOURCE;
  PRINT ",Length," LORA_RXPacketL;
  PRINT ",RXcount," LORA_RXpacketCount;
  PRINT ",CRCErrors," LORA_CRCerrorcount
  
  LORA_LVAR1 = 137 - LORA_PacketRSSI		'calculate true RSSI value
  PRINT "RSSI -" LORA_LVAR1 "dBm"
  LORA_LVAR1 = 137 - LORA_BackGroundRSSI		'calculate true RSSI value
  PRINT "Noise -" LORA_LVAR1 "dBm"
  
  if LORA_PacketSNR > 127 then
    LORA_LVAR1 =  (255 - LORA_PacketSNR) / 4
    LORA_LChar = "-"
  else
    LORA_LVAR1 = LORA_PacketSNR / 4
    LORA_LChar = "+"
  end if
  
  PRINT "PacketSNR " LORA_LChar LORA_LVAR1 "dB"
  
  
  
END SUB
  
  
SUB LORA.RXOFF
  'turns off receiver
  PRINT "LORA.RXOFF"
  LORA.Write(LORA_RegOpMode,&H89)                   'TX and RX to sleep, in direct mode
EnD SUB
  '************************
  'END LORA Receive Stuff
  '************************
  
